"""
------------------------------------------
frame_lib: Pythonic wrappers around FrameL
------------------------------------------
"""

# B. Swinkels, 2015

from __future__ import division
from builtins import str
from builtins import next
from builtins import range
from builtins import object
import ctypes
import logging
import os
from os import path
import sys
import numpy as np

import PyFd as fd
from virgotools.gpstime import start_stop_dur
from virgotools import cm_lib
from virgotools.pvt_utils import ascii, native_str

try:
    import virgotools.pvt_internal_defaults as defaults
except ImportError:
    import virgotools.pvt_external_defaults as defaults


__all__ = ['ChannelNotFound', 'FrameFile', 'getChannel', 'Frame',
           'FdIOError', 'FdIO', 'FrVect']


logger = logging.getLogger(__name__)


DEFAULT_CHANNEL_PREFIX = ascii(os.environ.get('CHANNEL_PREFIX', defaults.channel_prefix))
DEFAULT_FFL_PATH = ascii(os.environ.get('FFLPATH',defaults.ffl_path))  # same as gwpy


def expand_channel(chan):
    """convenience function that adds DEFAULT_CHANNEL_PREFIX to a channel name
     if it does not already have a prefix, and converts the result to ascii"""
    chan = ascii(chan)
    if chan[2:3] == b':':
        return chan
    return DEFAULT_CHANNEL_PREFIX + chan


def expand_ffl(ffl_file):
    """convenience function that allows to use 'trend' instead of
    '/virgoData/ffl/trend.ffl' """
    ffl_file = ascii(ffl_file)
    if ffl_file.endswith((b'.ffl', b'.gwf')):
        return ffl_file
    if not DEFAULT_FFL_PATH:
        raise NameError('"Provide a full path to an ffl or gwf file, or define an environmental variable FFLPATH or edit the ffl_path variable in %s/pvt_external_defaults.p' %(os.path.dirname(os.path.abspath(__file__))))
    return path.join(DEFAULT_FFL_PATH, ffl_file + b'.ffl')


# mapping from FrVect.type to attribute with corresponding data pointer and
# corresponding numpy dtype
# see Fr/matlab/frgetvectN.c
_FR_TYPES = {
    fd.FR_VECT_C: ('dataU', np.uint8),  # not sure if it should be signed or unsigned
    fd.FR_VECT_2S: ('dataS', np.int16),
    fd.FR_VECT_4S: ('dataI', np.int32),
    fd.FR_VECT_8S: ('dataL', np.int64),
    fd.FR_VECT_1U: ('dataU', np.uint8),
    fd.FR_VECT_2U: ('dataUS', np.uint16),
    fd.FR_VECT_4U: ('dataUI', np.uint32),
    fd.FR_VECT_8U: ('dataUL', np.uint64),
    fd.FR_VECT_4R: ('dataF', np.float32),
    fd.FR_VECT_8R: ('dataD', np.float64),
    # skip complex formats for now
}


# common to FrameFile.getChannel and Frame.getAdc, without scaling
def FrVect2array(p_vect):
    """convert FrVect to Numpy array"""
    vect = p_vect.contents
    attr, dtype = _FR_TYPES[vect.type]
    p_data = getattr(vect, attr)
    shape = tuple(vect.nx[i] for i in range(vect.nDim))
    # FIXME: for image, which have ndim==3, need to swap last two sizes

    # allocate numpy vector and copy data over. Sharing data with
    # np.ctypeslib.as_array might be faster, but that leaks memory
    data = np.empty(shape, dtype)
    ctypes.memmove(data.ctypes.data, p_data, data.nbytes)
    return data


# custom exception
class ChannelNotFound(Exception):
    def __init__(self, channel):
        self.channel = channel
        super(ChannelNotFound, self).__init__(channel)  # so that pickle works

    def __str__(self):
        return 'Could not find channel %s in frame!' % self.channel


def decode_sms(s):
    """splits SMS string and yields tuples (chan, value, units)"""
    it = iter(s.split())
    units = b'Counts.'
    for chan in it:
        if chan in (b'bias', b'slope'):
            raise NotImplementedError
        if chan == b'units':
            units = next(it)
            continue
        value = float(next(it))
        yield chan, value, units


class Frame(object):
    """
    Pythonic wrapper class around the FrameH structure.

    A Frame object gives access to all important properties, such as the
    gps-time, and allows to retrieve the values of channels as Numpy vectors. As
    a convenience, adc-channels and sms-values contained in a frame can also be
    obtained via attribute access. Frame objects are typically returned by
    :class:`FrameFile` or :class:`FdIO` objects.

    Parameters
    ----------
    p_frame
        ctypes pointer to FrameH structure

    Attributes
    ----------
    gps : int
        gps-time of the start of the frame
    dt : int
        length of the frame in seconds
    gap: int
        gap between this frame and the previous frame (obtained by e.g
        FrameFile.next or FdIO.get_frame) in seconds. When this value is not
        applicable (e.g. when asking a single frame, or for the first frame in a
        series) gap is set to -1. When receiving a contiguous series of frames,
        gap will be zero. In case of a missing frame, gap will have a positive
        value. This could be used e.g. to reset filters.

    Examples
    --------
    >>> frame.gps
    1144144190
    >>> frame.INJ_IMC_TRA_DC
    array([ 24.73620276,  24.73603786,  24.73647757, ...,  24.74477708,
            24.74785504,  24.75258191])
    >>> frame.NON_EXISTING
    [ ... Traceback ...]
    frame_lib.ChannelNotFound: Could not find channel V1:NON_EXISTING in frame!

    """
    def __init__(self, p_frame):
        self._frame = p_frame
        self.gap = -1

    def close(self):
        """Free memory of frame.

        To avoid memory leaks, this function needs to be called when the frame
        is no longer used (except when returning the frame using FdIO.put_frame).
        Instead of explicitly closing the frame, you can use a *with-statement*.

        Examples
        --------
        Explicit closing of frame:

        >>> ffl = FrameFile('raw')
        >>> frame = ffl.last_frame()
        >>> do_something(frame)
        >>> frame.close()

        More robust way of implicit closing a frame using a with statement:

        >>> ffl = FrameFile('raw')
        >>> with ffl.last_frame() as frame:  # frame gets closed after with block
        ...     do_something(frame)

        This will close the frame after the with statement even in case of
        Exceptions.
        """
        fd.FrameFree(self._frame)
        self._frame = None

    def _calculate_gap(self, last_gps):
        """
        utility function for setting the gap attribute, to be called by
        objects that provide several frames in series. Returns gps of end of
        frame.

        Example
        -------
        last_gps = None
        while True:
            frame = get_some_frame()
            last_gps = frame._calculate_gap(last_gps)
        """
        if last_gps is None:
            self.gap = -1
        else:
            self.gap = self.gps - last_gps
        return self.gps + self.dt

    # context manager protocol to allow with Frame: ...
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def gps(self):
        return int(self._frame.contents.GTimeS)

    @property
    def dt(self):
        return int(self._frame.contents.dt)

    def copy(self):
        """
        makes an empty copy of the current frame with the correct gps and
        other atributes. The new frame can be used by processes like PyALP to
        create frames that contain only new channels, which avoids issues with
        duplicate channels in the DAQ chain.

        Returns
        -------
        Frame
            empty copy of the current frame

        Example
        -------
        >>> in_frame = fdio.get_frame()
        >>> out_frame = in_frame.copy()
        >>> # add new channels to out_frame based on in_frame
        >>> in_frame.close()
        >>> fdio.put_frame(out_frame)
        """
        p_copy = fd.FrameHCopy(self._frame)
        assert p_copy, 'could not copy frame'

        return Frame(p_copy)

    def iter_adc(self):
        """
        Iterate over all FrAdcData structs inside the frame.

        Yields
        ------
        pointer to FrAdcData struct for all channels in the frame.

        Example
        -------
        >>> for adc in frame.iter_adc():
        ...     print('Channel %s with sample rate %i Hz' % (
        ...         adc.contents.name, adc.contents.sampleRate))

        """
        adc = self._frame.contents.rawData.contents.firstAdc
        while adc:
            yield adc
            adc = adc.contents.next

    def iter_sms(self):
        """Iterate over all FrSerData structs inside the frame."""
        ser = self._frame.contents.rawData.contents.firstSer
        while ser:
            yield ser
            ser = ser.contents.next

    def iter_sms_channels(self):
        """Iterate over all sms channels inside a frame."""
        for ser in self.iter_sms():
            head = ser.contents.name
            data = ser.contents.data
            fsamp = ser.contents.sampleRate
            for tail, value, units in decode_sms(data):
                yield b'%s_%s' % (head, tail), value, units, fsamp

    def get_adc_channel(self, channel):
        """lookup an adc channel on the frame

        Parameters
        ----------
        channel : str
            name of the adc channel to be looked up

        Returns
        -------
        ndarray
            vector with values of the channel

        Raises
        ------
        ChannelNotFound
            if the adc channel does not exist in the frame

        Example
        -------
        >>> raw = FrameFile('raw')
        >>> frame = raw.last_frame()
        >>> frame.get_adc_channel('V1:INJ_IMC_TRA_DC')
        array([ 23.99298497,  23.99303994,  23.99287504, ...,  23.99496366,
                23.99474381,  23.99457892])
        """
        channel = ascii(channel)
        # get pointer of correct data type
        p_adc_data = fd.FrAdcDataFind(self._frame, channel)
        if not p_adc_data:
            raise ChannelNotFound(channel)
        adc_data = p_adc_data.contents  # FrAdcData struct
        data = FrVect2array(adc_data.data)

        # apply scaling if necessary
        if adc_data.slope == 1.0 and adc_data.bias == 0.0:
            return data
        data = data.astype(np.float64, copy=False)  # convert in-place
        data *= adc_data.slope
        data += adc_data.bias
        return data

    # based on matlab_app.c/GetSmsChannel
    def get_sms_channel(self, channel):
        """lookup an sms value on the frame

        Parameters
        ----------
        channel : str
            name of the sms channel to be looked up

        Returns
        -------
        float
            value of the sms channel

        Raises
        ------
        ChannelNotFound
            if the sms channel does not exist in the frame

        Example
        -------
        >>> frame.get_sms_channel('V1:XX_YY_ZZ')
        1.23
        """
        channel = ascii(channel)
        parts = channel.split(b'_')
        for i in range(1, len(parts)):  # try a..b_c_d, a_b..c_d, a_b_c..d ...
            # FIXME: use FrSerDataGet0, but not yet wrapped in PyFd
            val = fd.FrSerDataGetValue(
                self._frame, b'_'.join(parts[:i]), b'_'.join(parts[i:]), np.NaN)
            if not np.isnan(val):
                return val
        raise ChannelNotFound(channel)

    def getChannel(self, channel, **kwargs):
        """
        getChannel(channel, mask=False, fill_value=np.nan)
        get channel data from a frame, similar to :func:`getChannel`, but
        using the current frame (so no need to pass a source), and for the
        complete duration of this frame (so no need to pass start and end time)

        """
        # first try to find Adc data, needs to be special cased for slope/bias
        p_adc_data = fd.FrAdcDataFind(self._frame, channel)  # FIXME ascii()
        if p_adc_data:
            adc_data = p_adc_data.contents  # FrAdcData struct
            slope = adc_data.slope
            bias = adc_data.bias
            p_vect = adc_data.data
        else:  # if not found try procdata and others
            p_vect = fd.FrameFindVect(self._frame, channel)
            slope = 1.0
            bias = 0.0

        # finally try sms data, will raise ChannelNotFound if not found
        if not p_vect:  # FIXME: this does not return a FrVect
            return self.get_sms_channel(channel)

        vect = FrVect(p_vect, **kwargs)

        if slope == 1.0 and bias == 0.0:
            return vect
        vect.data = vect.data.astype(np.float64, copy=False)  # convert in-place
        vect.data *= slope
        vect.data += bias
        return vect

    def __getattr__(self, channel):
        """
        convenience method for looking up channels on a frame object.

        This method allows doing ``B1 = frame.LSC_NE_CORR`` instead of
        ``B1 = frame.get_adc_data('V1:LSC_NE_CORR')``. A prefix ``V1:`` is
        added, after which the attribute is looked up as a channel using
        :func:`get_adc_channel`, and if that fails using
        :func:`get_sms_channel`.

        Parameters
        ----------
        channel : str
            name of the channel to be looked up

        Raises
        ------
        ChannelNotFound
            if the attribute is not a valid channel

        Example
        -------
        >>> x = frame.INJ_IMC_TRA_DC
        """
        channel = ascii(channel)
        channel = expand_channel(channel)

        try:
            return self.get_adc_channel(channel)
        except ChannelNotFound:
            try:
                return self.get_sms_channel(channel)
            except ChannelNotFound:
                raise ChannelNotFound(channel)

    # based on matlap_app.c/UpdateOutputs
    def write_sms_prefix(self, prefix, chans_vals):
        """writes channel/value pairs to sms with common prefix
        chans_vals is an iterable of chan/value pairs, or a dict"""

        prefix = ascii(prefix)
        prefix = expand_channel(prefix)

        if isinstance(chans_vals, dict):
            it = chans_vals.items()
        else:
            it = iter(chans_vals)

        data = []
        for chan, value in it:
            chan = ascii(chan)
            try:
                value = float(value)
            except:
                raise ValueError('SMS value must be a number')
            data += [chan, b'%r' % value]
        assert data, 'must write at least one channel'
        data = b' '.join(data)
        serData = fd.FrSerDataNew(self._frame, prefix, self.gps, data,
                                  1.0 / self.dt)
        assert serData, 'FrSerDataNew failed'

    def write_adc_channel(self, chan, vect):  #fixme change to add_adc_channel?
        chan = ascii(chan)
        assert vect.dtype == np.float32  # FIXME: use np.asarray/np.require, default dtype=float32?
        p_adc_data = fd.FrAdcDataNew(self._frame, chan, len(vect) / self.dt,
                                     len(vect), -32)  # see matlab_app
        if not p_adc_data:
            raise FdIOError('FrAdcDataNew failed for channel %s' % chan)
        p_data = p_adc_data.contents.data.contents.dataF
        ctypes.memmove(p_data, vect.ctypes.data, vect.nbytes)  # best way??


class FrVect(object):
    """
    class used for results of :meth:`FrameFile.getChannel` and :func:`getChannel`

    Attributes
    ----------
    name : str
        channel name
    data : numpy.ndarray
        vector with content of channel
    unit : str
        units of the data
    gps : float
        gps-time of the start of the vector
    time : numpy.ndarray
        vector with the time offsets with respect to :attr:`gps` corresponding
        to the data samples, which can e.g. be used for plotting
    dt : float
        sample time
    fsample : float
        sample frequency
    missing : numpy.ndarray or None
        only present in case mask=True was used in getChannel. This argument is
        either None, or a vector that is zero for missing samples

    """
    # creates a new vector object given a pointer to a FrVect structure
    # this now no longer stores the vector, and freeing the vector is the
    # responsibility of the caller
    def __init__(self, p_vect, mask=False, fill_value=np.nan):
        self.name = native_str(p_vect.contents.name)
        self.dt = p_vect.contents.dx.contents.value
        self.unit = native_str(p_vect.contents.unitY)
        self._dx = p_vect.contents.startX.contents.value
        self.gps = p_vect.contents.GTime

        self.data = FrVect2array(p_vect)

        # handle missing data
        miss = p_vect.contents.next
        if miss:
            miss = FrVect2array(miss)
            if mask:
                self.missing = miss
            else:
                self.data[miss == 0] = fill_value
        else:
            if mask:
                self.missing = None

    @property
    def time(self):
        # see plhs[1] in Fr/matlab/frgetvectN
        time = np.arange(self.data.shape[0], dtype=np.float64)
        if self.dt != 1:
            time *= self.dt
        if self._dx:
            time += self._dx
        return time

    @property
    def fsample(self):
        f =  1.0 / self.dt
        rf = round(f)
        # fix floating point inaccuracies, e.g. 1 / 1-e5 != 100000
        if np.isclose(f, rf, atol=0, rtol=1e-10):
            f = rf
        return f

    # context manager protocol no longer needed, keep for backwards compatibility
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass  # used to free p_vect, no longer needed


class FrameFile(object):
    """
    Pythonic wrapper around FrFile to open frame files. The objects can be
    used as a context manager, so that they get automatically closed after
    use (using the with statement) and can be iterated over to obtain
    subsequent frames (using a for loop). :class:`Frame` objects can be
    extracted in several different ways, which should always be closed after
    use (using a with statement or explicitly using :meth:`Frame.close`).

    Parameters
    ----------
    fname : str
        name of a .ffl or .gwf file. The standard files in /virgoData/ffl can be
        accessed using the shortcuts 'trend', 'raw', ...

    Attributes
    ----------
    gps_start
        gps time of start of frame file
    gps_end
        gps time of end of frame file

    Example
    -------
    >>> with FrameFile('raw') as ffl:
    ...     for frame in ffl:
    ...             # do something with frame
    ...             frame.close()
    ...     with ffl.first_frame() as frame:
    ...         # do something with frame
    ...     with ffl.last_frame() as frame:
    ...         # do something with frame
    ...     with ffl.get_frame(gps) as frame:
    ...         # do something with frame
    """
    def __init__(self, fname):
        self.fname = expand_ffl(fname)
        self._FrFile = fd.FrFileINew(self.fname)
        if not self._FrFile:
            msg = 'could not open file %s' % fname
            raise IOError(msg)

        self._last_gps = None  # for calculating Frame.gap

    def close(self):
        fd.FrFileIEnd(self._FrFile)
        self._FrFile = None

    # context manager protocol
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    # iterator protocol, iterating over a FrameFile will yield subsequent frames
    def __iter__(self):
        return self

    def __next__(self):
        """returns the next frame. User should do FrameFree when done"""
        frame = fd.FrameRead(self._FrFile)
        if not frame:
            raise StopIteration
        frame = Frame(frame)
        self._last_gps = frame._calculate_gap(self._last_gps)
        return frame

    def seek(self, gps):
        """sets the internal pointer of a FrameFile to the frame containing the
        given gps. If the gps falls in a gap, the next frame will be pointed to.
        The intended use is set the starting gps before iterating over the
        FrameFile object using a for-loop.

        Parameters
        ----------
        gps : float
            gps time of the frame to point to

        Raises
        ------
        ValueError
            If the gps is out of range

        Example
        -------
        >>> ffl = FrameFile('raw')
        >>> g = now_gps() - 300
        >>> ffl.seek(g)
        >>> for frame in ffl:
        ...     with frame as f:
        ...         print('got frame with gps =', f.gps)
        """
        result = fd.FrFileISetTime(self._FrFile, gps)
        if result < 0:
            raise ValueError('Could no seek to gps = %i' %  gps)
        self._last_gps = None  # reset gap calculation

    def get_frame(self, gps):
        """return the frame that contains the given gps time. This means that
        the start gps of the frame can be lower than the requested gps time, but
        the difference should always be smaller than the length of one frame.
        The user should do frame.close() or do
        "with ffl.get_frame(gps) as frame:" to properly close the frame after
        use.

        Parameters
        ----------
        gps : float
            gps time at which to get a frame

        Returns
        -------
        Frame
            frame at requested gps time

        Example
        -------
        >>> with FrameFile('raw') as ffl:
        ...     with ffl.get_frame(1228550185) as frame:
        ...         print frame.gps
        1228550180
        """
        p_frame = fd.FrameReadT(self._FrFile, gps)
        if not p_frame:
            raise ValueError('Could not get frame at gps = %i' % gps)
        frame = Frame(p_frame)
        self._last_gps = frame._calculate_gap(self._last_gps)
        return frame

    def first_frame(self):
        """returns first frame in ffl file as Frame object. User should do
        frame.close() or use a "with ffl.first_frame() as frame:" to properly
        close the frame after use.

        Returns
        -------
        Frame
            first frame in the frame file
        """
        fd.FrFileIRewind(self._FrFile)
        frame = Frame(fd.FrameRead(self._FrFile))
        self._last_gps = frame._calculate_gap(None)
        return frame

    def last_frame(self):
        """returns last frame in ffl file as Frame object. User should do
        frame.close() or use a "with ffl.last_frame() as frame:" to properly
        close the frame after use

        Returns
        -------
        Frame
            last frame in the frame file
        """
        frame = Frame(fd.FrameReadT(self._FrFile, self.gps_end - 0.1))
        self._last_gps = frame._calculate_gap(None)
        return frame

    @property
    def gps_start(self):
        return fd.FrFileITStart(self._FrFile)

    @property
    def gps_end(self):
        return fd.FrFileITEnd(self._FrFile)

    # inspired by Fr/matlab/frgetvectN
    def getChannel(self, channel, gstart, gstop_or_dur, **kwargs):
        """
        getChannel(channel, gstart, gstop_or_dur, mask=False, fill_value=np.nan)
        same as :func:`getChannel`, but for this FrameFile object (so without
        the source argument), and without adding any V1: prefix.

        Example
        -------
        >>> with FrameFile('raw') as ffl:
        ...     x = ffl.getChannel('V1:INJ_IMC_TRA_DC', now_gps() - 3600, 10)
        """
        channel = ascii(channel)
        gstart, _, dur = start_stop_dur(gstart, gstop_or_dur)
        p_vect = fd.FrFileIGetVectDN(self._FrFile, channel, gstart, dur)
        if not p_vect:
            raise ChannelNotFound(channel)
        vect = FrVect(p_vect, **kwargs)
        fd.FrVectFree(p_vect)
        return vect


# equivalent of matlab getChannel
def getChannel(*args, **kwargs):
    """
    getChannel([source,] channel, gstart, gstop_or_dur, mask=False, fill_value=np.nan)
    read channel data from frame file

    Parameters
    ----------
    source : str
        frame file, either a full path to a ffl or gwf file, or an abbreviation
        like 'raw', 'trend', which are looked up in the stancard location. If
        omitted, defaults to 'raw', but this default value is deprecated.
    channel : str
        channel name. If missing, a prefix 'V1:' is added.
    gstart : float
        starting gps time
    gstop_or_dur : float
        either a duration in seconds or the stop gps time
    mask : bool
        if mask is False, missing samples will be replaced by fill_value. If it
        is True, the returned FrVect will have an attribute missing, which is a
        mask vector that is zero for missing samples.
    fill_value : float
        value that is used for missing samples if mask is False

    Returns
    -------
    FrVect
        object with the content of the channel, units, time, ...

    Example
    -------
    >>> x = getChannel('raw', 'INJ_IMC_TRA_DC', now_gps() - 3600, 10)
    >>> import matplotlib.pyplot as plt
    >>> plt.plot(x.time, x.data)
    >>> plt.show()
    """
    if len(args) == 3:  # FIXME deprecate implicit raw?
        source = b'raw'
        channel, gstart, gstop_or_dur = args
    elif len(args) == 4:
        source, channel, gstart, gstop_or_dur = args
    else:
        raise ValueError('getChannel needs 3 or 4 inputs')

    channel = ascii(channel)
    source = ascii(source)

    channel = expand_channel(channel)

    with FrameFile(source) as ffl:
        return ffl.getChannel(channel, gstart, gstop_or_dur, **kwargs)


# see http://stackoverflow.com/questions/3494598
def _string_array(L):
    """covert list of strings to array of char pointers"""
    arr = (ctypes.c_char_p * len(L))()
    arr[:] = L
    return arr


# custom exception
class FdIOError(Exception):
    pass


class FdIO(object):
    """
    Pythonic wrapper around FdIO objects

    Parameters
    ----------
    argv : list
        command line arguments. If omitted, :py:data:`sys.argv` is used.

    Raises
    ------
    FdIOError
        If FdIO cannot be initizialized

    Example
    -------

    Minimal working example, requires a proper .cfg file

    ::

        import sys
        fdio = FdIO(sys.argv)
        while not fd.CfgFinished():
            try:
                frame = fdio.get_frame()
            except FdIOError:  # happens a lot when waiting for cm
                cm.CmMessageWaitWithTimeout(0.1)
                continue
            print('GPS of received frame:', frame.gps)
            fdio.put_frame(frame)

    See the PyALP or TangoDAQSmsBridge source codes for more complete examples
    """
    def __init__(self, argv=None):
        if argv is None:
            argv = sys.argv

        argv = [ascii(arg) for arg in argv]

        logger.info('Init FdIO using config file %s', argv[1])
        if len(argv) > 2:
            cm_name = argv[2]
            logger.info('Using explicit cm-name %s', cm_name)
        else:
            cm_name = path.splitext(path.basename(argv[1]))[0]
            logger.info('Using implicit cm-name %s from config file', cm_name)

        cm_name = ascii(cm_name)

        if cm_lib._my_cm_name is not None:
            msg = 'Starting FdIO, but already have a cm-name %s' % cm_lib._my_cm_name
            raise FdIOError(msg)
        cm_lib._my_cm_name = cm_name

        argv = _string_array(argv)
        self._fdio = fd.FdIONew(len(argv), argv)
        if not self._fdio:
            raise FdIOError('FdIONew: failed init with %s' % argv[1])

        logger.info('Parsing config file')
        fd.FdIOParseAndIni(self._fdio)
        # TODO: how to check if parsing failed? check source

    def get_frame(self):
        """
        Get a frame via the Fd system

        Note that when done using the frame, it should not be closed but should
        be returned using :meth:`put_frame`

        Raises
        ------
        FdIOError
            when no frame is obtained

        Returns
        -------
        frame : Frame

        """
        p_frame = fd.FdIOGetFrame(self._fdio)
        if not p_frame:  # TODO: return None?
            raise FdIOError('FdIOGetFrame: could not get frame')
        return Frame(p_frame)

    def put_frame(self, frame):
        """
        Return a frame to the Fd system.

        Parameters
        ----------
        frame : Frame
            frame object to be returned
        """
        fd.FdIOPutFrame(self._fdio, frame._frame)
        frame._frame = None

